"""Utilities for selecting Kiln tools to expose via the MCP server."""

from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Sequence

from kiln_ai.datamodel.external_tool_server import ToolServerType
from kiln_ai.datamodel.project import Project
from kiln_ai.datamodel.tool_id import build_kiln_task_tool_id, build_rag_tool_id
from kiln_ai.tools.base_tool import KilnToolInterface
from kiln_ai.tools.kiln_task_tool import KilnTaskTool
from kiln_ai.tools.rag_tools import RagTool

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class ToolResolution:
    """Represents a tool resolved from the Kiln project."""

    tool_id: str
    tool: KilnToolInterface


def collect_project_tools(
    project: Project,
    allowed_tool_ids: Sequence[str] | None = None,
) -> list[ToolResolution]:
    """Collect the Kiln tools that should be exposed via MCP.

    Args:
        project: Project containing potential tools.
        allowed_tool_ids: Optional sequence of tool IDs to include. If provided,
            only matching tools will be returned.

    Returns:
        A list of :class:`ToolResolution` objects describing the selected tools.

    Raises:
        ValueError: If ``allowed_tool_ids`` contains IDs that do not resolve to
            available tools.
    """

    allowed_set = set(allowed_tool_ids or [])
    missing_ids = set(allowed_set)
    resolutions: list[ToolResolution] = []

    for rag_config in project.rag_configs(readonly=True):
        if rag_config.is_archived:
            logger.debug(
                "Skipping archived RAG config %s",
                getattr(rag_config, "id", "<unknown>"),
            )
            continue

        tool_id = build_rag_tool_id(rag_config.id)

        if allowed_set and tool_id not in allowed_set:
            logger.debug(
                "Skipping tool %s because it is not in the allowed list", tool_id
            )
            continue

        tool = RagTool(tool_id, rag_config)
        resolutions.append(ToolResolution(tool_id=tool_id, tool=tool))
        missing_ids.discard(tool_id)

    for server in project.external_tool_servers(readonly=True):
        if server.type != ToolServerType.kiln_task:
            continue

        tool_id = build_kiln_task_tool_id(server.id)

        if allowed_set and tool_id not in allowed_set:
            logger.debug(
                "Skipping tool %s because it is not in the allowed list", tool_id
            )
            continue

        if project.id is None:
            raise ValueError("Project ID is required")

        tool = KilnTaskTool(project.id, tool_id, server)
        resolutions.append(ToolResolution(tool_id=tool_id, tool=tool))
        missing_ids.discard(tool_id)

    if missing_ids:
        raise ValueError(
            "Requested tool IDs were not found or are archived: "
            + ", ".join(sorted(missing_ids))
        )

    return resolutions



"""Shared utility functions for spec and eval creation.

This module contains helper functions for creating specs and their associated evals.
These functions are used by both the core spec API and the desktop copilot API.
"""

from kiln_ai.datamodel.datamodel_enums import TaskOutputRatingType
from kiln_ai.datamodel.eval import EvalDataType, EvalOutputScore, EvalTemplateId
from kiln_ai.datamodel.spec_properties import SpecType


def spec_eval_output_score(spec_name: str) -> EvalOutputScore:
    """Create an EvalOutputScore for a spec."""
    return EvalOutputScore(
        name=spec_name,
        type=TaskOutputRatingType.pass_fail,
        instruction=f"Evaluate if the model's behaviour meets the spec: {spec_name}.",
    )


def spec_eval_data_type(
    spec_type: SpecType, evaluate_full_trace: bool = False
) -> EvalDataType:
    """Determine the eval data type for a spec."""
    if spec_type == SpecType.reference_answer_accuracy:
        return EvalDataType.reference_answer

    if evaluate_full_trace:
        return EvalDataType.full_trace
    else:
        return EvalDataType.final_answer


def spec_eval_template(spec_type: SpecType) -> EvalTemplateId | None:
    """Get the eval template for a spec type."""
    match spec_type:
        case SpecType.appropriate_tool_use:
            return EvalTemplateId.tool_call
        case SpecType.reference_answer_accuracy:
            return EvalTemplateId.rag
        case SpecType.factual_correctness:
            return EvalTemplateId.factual_correctness
        case SpecType.toxicity:
            return EvalTemplateId.toxicity
        case SpecType.bias:
            return EvalTemplateId.bias
        case SpecType.maliciousness:
            return EvalTemplateId.maliciousness
        case SpecType.jailbreak:
            return EvalTemplateId.jailbreak
        case SpecType.issue:
            return EvalTemplateId.issue
        case SpecType.desired_behaviour:
            return EvalTemplateId.desired_behaviour
        case (
            SpecType.tone
            | SpecType.formatting
            | SpecType.localization
            | SpecType.hallucinations
            | SpecType.completeness
            | SpecType.nsfw
            | SpecType.taboo
            | SpecType.prompt_leakage
        ):
            return None


def generate_spec_eval_tags(spec_name: str) -> tuple[str, str, str]:
    """Generate eval, train, and golden tags for a spec.

    Args:
        spec_name: The name of the spec

    Returns:
        Tuple of (eval_tag, train_tag, golden_tag)
    """
    tag_suffix = spec_name.lower().replace(" ", "_")
    eval_tag = f"eval_{tag_suffix}"
    train_tag = f"train_{tag_suffix}"
    golden_tag = f"eval_golden_{tag_suffix}"
    return eval_tag, train_tag, golden_tag


def generate_spec_eval_filter_ids(
    eval_tag: str, train_tag: str, golden_tag: str
) -> tuple[str, str, str]:
    """Generate filter IDs for eval set, train set, and eval configs.

    Args:
        eval_tag: The eval dataset tag
        train_tag: The train dataset tag
        golden_tag: The golden dataset tag

    Returns:
        Tuple of (eval_set_filter_id, train_set_filter_id, eval_configs_filter_id)
    """
    return f"tag::{eval_tag}", f"tag::{train_tag}", f"tag::{golden_tag}"



from unittest.mock import patch

import pytest
from fastapi import FastAPI
from fastapi.testclient import TestClient
from kiln_ai.datamodel import Project, Prompt, PromptGenerators, Task

from kiln_server.custom_errors import connect_custom_errors
from kiln_server.prompt_api import _prompt_generators, connect_prompt_api


@pytest.fixture
def app():
    app = FastAPI()
    connect_prompt_api(app)
    connect_custom_errors(app)
    return app


@pytest.fixture
def client(app):
    return TestClient(app)


@pytest.fixture
def project_and_task(tmp_path):
    project_path = tmp_path / "test_project" / "project.kiln"
    project_path.parent.mkdir()

    project = Project(name="Test Project", path=str(project_path))
    project.save_to_file()
    task = Task(
        name="Test Task",
        instruction="This is a test instruction",
        description="This is a test task",
        parent=project,
    )
    task.save_to_file()

    return project, task


def test_create_prompt_success(client, project_and_task):
    project, task = project_and_task

    prompt_data = {
        "name": "Test Prompt",
        "prompt": "This is a test prompt",
        "description": "This is a test prompt description",
        "chain_of_thought_instructions": "Think step by step, explaining your reasoning.",
    }

    with patch("kiln_server.prompt_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.post(
            f"/api/projects/{project.id}/task/{task.id}/prompt", json=prompt_data
        )

    assert response.status_code == 200
    res = response.json()
    assert res["name"] == "Test Prompt"
    assert res["description"] == "This is a test prompt description"
    assert res["prompt"] == "This is a test prompt"

    # Check that the prompt was saved to the task/file
    prompts = task.prompts()
    assert len(prompts) == 1
    assert prompts[0].name == "Test Prompt"
    assert prompts[0].prompt == "This is a test prompt"
    assert (
        prompts[0].chain_of_thought_instructions
        == "Think step by step, explaining your reasoning."
    )


def test_create_prompt_task_not_found(client):
    prompt_data = {
        "name": "Test Prompt",
        "prompt": "This is a test prompt",
    }

    response = client.post(
        "/api/projects/project-id/task/fake-task-id/prompt", json=prompt_data
    )
    assert response.status_code == 404


def test_get_prompts_success(client, project_and_task):
    project, task = project_and_task

    test_prompt = Prompt(
        name="Test Prompt", prompt="This is a test prompt", parent=task
    )
    test_prompt.save_to_file()

    with patch("kiln_server.prompt_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.get(f"/api/projects/{project.id}/task/{task.id}/prompts")

    assert response.status_code == 200
    res = response.json()
    assert isinstance(res, dict)
    assert "generators" in res
    assert "prompts" in res
    assert len(res["generators"]) > 0  # Should have our predefined generators
    assert len(res["prompts"]) == 1
    assert res["prompts"][0]["name"] == "Test Prompt"


def test_get_prompts_task_not_found(client):
    response = client.get("/api/projects/project-id/task/fake-task-id/prompts")
    assert response.status_code == 404


def test_prompt_generators_content():
    """Test that our predefined prompt generators have the expected structure"""
    from kiln_server.prompt_api import _prompt_generators

    # Test a few key generators
    basic = next(g for g in _prompt_generators if g.id == "simple_prompt_builder")
    assert basic.chain_of_thought is False
    assert "zero-shot" in basic.description.lower()

    cot = next(
        g
        for g in _prompt_generators
        if g.id == "simple_chain_of_thought_prompt_builder"
    )
    assert cot.chain_of_thought is True
    assert "Chain of Thought" in cot.name


# Check our nice UI list with descriptions covers all our generators
def test_all_ids_are_covered():
    generators = [e.value for e in PromptGenerators]
    api_list = [g.id for g in _prompt_generators]

    assert set(api_list) == set(generators)


def test_update_prompt_success(client, project_and_task):
    project, task = project_and_task

    # First create a prompt
    test_prompt = Prompt(
        name="Original Name",
        prompt="This is a test prompt",
        description="Original description",
        parent=task,
    )
    test_prompt.save_to_file()

    # Prepare update data
    update_data = {"name": "Updated Name", "description": "Updated description"}

    with patch("kiln_server.prompt_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.patch(
            f"/api/projects/{project.id}/tasks/{task.id}/prompts/id::{test_prompt.id}",
            json=update_data,
        )

    assert response.status_code == 200
    res = response.json()
    assert res["name"] == "Updated Name"
    assert res["description"] == "Updated description"

    # Verify the prompt was updated in the task/file
    updated_prompt = next((p for p in task.prompts() if p.id == test_prompt.id), None)
    assert updated_prompt is not None
    assert updated_prompt.name == "Updated Name"
    assert updated_prompt.description == "Updated description"
    # Ensure the prompt content wasn't changed
    assert updated_prompt.prompt == "This is a test prompt"


def test_update_prompt_not_found(client, project_and_task):
    project, task = project_and_task

    update_data = {"name": "Updated Name", "description": "Updated description"}

    with patch("kiln_server.prompt_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.patch(
            f"/api/projects/{project.id}/tasks/{task.id}/prompts/id::nonexistent_id",
            json=update_data,
        )

    assert response.status_code == 404
    assert "Prompt not found" in response.json()["message"]


def test_update_prompt_unsupported_type(client, project_and_task):
    project, task = project_and_task

    update_data = {"name": "Updated Name", "description": "Updated description"}

    with patch("kiln_server.prompt_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.patch(
            f"/api/projects/{project.id}/tasks/{task.id}/prompts/fine_tune_prompt::some_id",
            json=update_data,
        )

    assert response.status_code == 400
    assert "Only custom prompts can be updated" in response.json()["message"]


def test_delete_prompt_success(client, project_and_task):
    project, task = project_and_task

    # First create a prompt
    test_prompt = Prompt(
        name="Test Prompt",
        prompt="This is a test prompt",
        description="This is a test prompt description",
        parent=task,
    )
    test_prompt.save_to_file()

    # Verify the prompt exists
    prompts_before = task.prompts()
    assert len(prompts_before) == 1
    assert prompts_before[0].id == test_prompt.id

    # Store the path to check if file is deleted
    prompt_path = test_prompt.path

    with patch("kiln_server.prompt_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.delete(
            f"/api/projects/{project.id}/tasks/{task.id}/prompts/id::{test_prompt.id}"
        )

    assert response.status_code == 200

    # Verify the prompt was deleted from the task
    prompts_after = task.prompts()
    assert len(prompts_after) == 0

    # Verify the file was actually deleted, including the parent directory
    assert not prompt_path.exists()
    assert not prompt_path.parent.exists()



import logging
from typing import List

from fastapi import FastAPI, HTTPException
from kiln_ai.datamodel.basemodel import FilenameString
from kiln_ai.datamodel.datamodel_enums import Priority
from kiln_ai.datamodel.eval import Eval
from kiln_ai.datamodel.spec import Spec, SpecStatus, TaskSample
from kiln_ai.datamodel.spec_properties import SpecProperties
from pydantic import BaseModel, Field

from kiln_server.task_api import task_from_id
from kiln_server.utils.spec_utils import (
    generate_spec_eval_filter_ids,
    generate_spec_eval_tags,
    spec_eval_data_type,
    spec_eval_output_score,
    spec_eval_template,
)

logger = logging.getLogger(__name__)


class UpdateSpecRequest(BaseModel):
    name: FilenameString | None = None
    definition: str | None = None
    properties: SpecProperties | None = Field(
        default=None,
        discriminator="spec_type",
    )
    priority: Priority | None = None
    status: SpecStatus | None = None
    tags: List[str] | None = None


def spec_from_id(project_id: str, task_id: str, spec_id: str) -> Spec:
    parent_task = task_from_id(project_id, task_id)
    spec = Spec.from_id_and_parent_path(spec_id, parent_task.path)
    if spec:
        return spec

    raise HTTPException(
        status_code=404,
        detail=f"Spec not found. ID: {spec_id}",
    )


class SpecCreationRequest(BaseModel):
    name: FilenameString
    definition: str
    properties: SpecProperties = Field(
        discriminator="spec_type",
    )
    priority: Priority = Field(default=Priority.p1)
    status: SpecStatus = Field(default=SpecStatus.active)
    tags: List[str] = Field(default_factory=list)
    evaluate_full_trace: bool = Field(default=False)
    task_sample: TaskSample | None = None


def connect_spec_api(app: FastAPI):
    @app.post("/api/projects/{project_id}/tasks/{task_id}/spec")
    async def create_spec(
        project_id: str, task_id: str, spec_data: SpecCreationRequest
    ) -> Spec:
        task = task_from_id(project_id, task_id)

        spec_type = spec_data.properties["spec_type"]

        eval_tag, train_tag, golden_tag = generate_spec_eval_tags(spec_data.name)
        eval_set_filter_id, train_set_filter_id, eval_configs_filter_id = (
            generate_spec_eval_filter_ids(eval_tag, train_tag, golden_tag)
        )

        template = spec_eval_template(spec_type)
        output_scores = [spec_eval_output_score(spec_data.name)]
        evaluation_data_type = spec_eval_data_type(
            spec_type, spec_data.evaluate_full_trace
        )

        eval_model = Eval(
            parent=task,
            name=spec_data.name,
            description=None,
            template=template,
            output_scores=output_scores,
            eval_set_filter_id=eval_set_filter_id,
            train_set_filter_id=train_set_filter_id,
            eval_configs_filter_id=eval_configs_filter_id,
            template_properties=None,
            evaluation_data_type=evaluation_data_type,
        )

        spec = Spec(
            parent=task,
            name=spec_data.name,
            definition=spec_data.definition,
            properties=spec_data.properties,
            priority=spec_data.priority,
            status=spec_data.status,
            tags=spec_data.tags,
            eval_id=eval_model.id,
            task_sample=spec_data.task_sample,
        )

        eval_model.save_to_file()
        try:
            spec.save_to_file()
        except Exception:
            eval_model.delete()
            raise

        return spec

    @app.get("/api/projects/{project_id}/tasks/{task_id}/specs")
    async def get_specs(project_id: str, task_id: str) -> List[Spec]:
        parent_task = task_from_id(project_id, task_id)
        return parent_task.specs(readonly=True)

    @app.get("/api/projects/{project_id}/tasks/{task_id}/specs/{spec_id}")
    async def get_spec(project_id: str, task_id: str, spec_id: str) -> Spec:
        return spec_from_id(project_id, task_id, spec_id)

    @app.patch("/api/projects/{project_id}/tasks/{task_id}/specs/{spec_id}")
    async def update_spec(
        project_id: str, task_id: str, spec_id: str, request: UpdateSpecRequest
    ) -> Spec:
        spec = spec_from_id(project_id, task_id, spec_id)

        # Update all provided fields
        if request.name is not None:
            spec.name = request.name
        if request.definition is not None:
            spec.definition = request.definition
        if request.properties is not None:
            spec.properties = request.properties
        if request.priority is not None:
            spec.priority = request.priority
        if request.status is not None:
            spec.status = request.status
        if request.tags is not None:
            spec.tags = request.tags

        spec.save_to_file()
        return spec

    @app.delete("/api/projects/{project_id}/tasks/{task_id}/specs/{spec_id}")
    async def delete_spec(project_id: str, task_id: str, spec_id: str) -> None:
        spec = spec_from_id(project_id, task_id, spec_id)

        # Delete associated eval if it exists
        if spec.eval_id:
            parent_task = task_from_id(project_id, task_id)
            eval = Eval.from_id_and_parent_path(spec.eval_id, parent_task.path)
            if eval:
                eval.delete()

        spec.delete()



"""Tests for kiln_server/utils/spec_utils.py."""

import pytest
from kiln_ai.datamodel.datamodel_enums import TaskOutputRatingType
from kiln_ai.datamodel.eval import EvalDataType, EvalTemplateId
from kiln_ai.datamodel.spec_properties import SpecType
from kiln_server.utils.spec_utils import (
    generate_spec_eval_filter_ids,
    generate_spec_eval_tags,
    spec_eval_data_type,
    spec_eval_output_score,
    spec_eval_template,
)


class TestSpecEvalOutputScore:
    def test_creates_eval_output_score_with_correct_name(self):
        score = spec_eval_output_score("Test Spec")
        assert score.name == "Test Spec"

    def test_creates_eval_output_score_with_pass_fail_type(self):
        score = spec_eval_output_score("Test Spec")
        assert score.type == TaskOutputRatingType.pass_fail

    def test_creates_eval_output_score_with_correct_instruction(self):
        score = spec_eval_output_score("Test Spec")
        assert "Test Spec" in score.instruction
        assert "meets the spec" in score.instruction


class TestSpecEvalDataType:
    def test_reference_answer_accuracy_returns_reference_answer(self):
        result = spec_eval_data_type(SpecType.reference_answer_accuracy)
        assert result == EvalDataType.reference_answer

    def test_reference_answer_accuracy_ignores_evaluate_full_trace(self):
        result = spec_eval_data_type(
            SpecType.reference_answer_accuracy, evaluate_full_trace=True
        )
        assert result == EvalDataType.reference_answer

    def test_other_types_default_to_final_answer(self):
        for spec_type in [
            SpecType.desired_behaviour,
            SpecType.issue,
            SpecType.tone,
            SpecType.toxicity,
        ]:
            result = spec_eval_data_type(spec_type)
            assert result == EvalDataType.final_answer

    def test_evaluate_full_trace_returns_full_trace(self):
        result = spec_eval_data_type(
            SpecType.desired_behaviour, evaluate_full_trace=True
        )
        assert result == EvalDataType.full_trace

    def test_evaluate_full_trace_false_returns_final_answer(self):
        result = spec_eval_data_type(
            SpecType.desired_behaviour, evaluate_full_trace=False
        )
        assert result == EvalDataType.final_answer


class TestSpecEvalTemplate:
    @pytest.mark.parametrize(
        "spec_type,expected_template",
        [
            (SpecType.appropriate_tool_use, EvalTemplateId.tool_call),
            (SpecType.reference_answer_accuracy, EvalTemplateId.rag),
            (SpecType.factual_correctness, EvalTemplateId.factual_correctness),
            (SpecType.toxicity, EvalTemplateId.toxicity),
            (SpecType.bias, EvalTemplateId.bias),
            (SpecType.maliciousness, EvalTemplateId.maliciousness),
            (SpecType.jailbreak, EvalTemplateId.jailbreak),
            (SpecType.issue, EvalTemplateId.issue),
            (SpecType.desired_behaviour, EvalTemplateId.desired_behaviour),
        ],
    )
    def test_returns_correct_template_for_spec_type(self, spec_type, expected_template):
        result = spec_eval_template(spec_type)
        assert result == expected_template

    @pytest.mark.parametrize(
        "spec_type",
        [
            SpecType.tone,
            SpecType.formatting,
            SpecType.localization,
            SpecType.hallucinations,
            SpecType.completeness,
            SpecType.nsfw,
            SpecType.taboo,
            SpecType.prompt_leakage,
        ],
    )
    def test_returns_none_for_spec_types_without_template(self, spec_type):
        result = spec_eval_template(spec_type)
        assert result is None


class TestGenerateSpecEvalTags:
    def test_generates_correct_tags_for_simple_name(self):
        eval_tag, train_tag, golden_tag = generate_spec_eval_tags("Test Spec")
        assert eval_tag == "eval_test_spec"
        assert train_tag == "train_test_spec"
        assert golden_tag == "eval_golden_test_spec"

    def test_handles_already_lowercase_name(self):
        eval_tag, train_tag, golden_tag = generate_spec_eval_tags("my spec")
        assert eval_tag == "eval_my_spec"
        assert train_tag == "train_my_spec"
        assert golden_tag == "eval_golden_my_spec"

    def test_handles_uppercase_name(self):
        eval_tag, train_tag, golden_tag = generate_spec_eval_tags("MY SPEC")
        assert eval_tag == "eval_my_spec"
        assert train_tag == "train_my_spec"
        assert golden_tag == "eval_golden_my_spec"

    def test_handles_single_word_name(self):
        eval_tag, train_tag, golden_tag = generate_spec_eval_tags("Toxicity")
        assert eval_tag == "eval_toxicity"
        assert train_tag == "train_toxicity"
        assert golden_tag == "eval_golden_toxicity"

    def test_train_tag_uses_train_prefix(self):
        """Train tag should use 'train_' prefix, not 'eval_train_'."""
        _, train_tag, _ = generate_spec_eval_tags("Test")
        assert train_tag.startswith("train_")
        assert not train_tag.startswith("eval_train_")


class TestGenerateSpecEvalFilterIds:
    def test_generates_correct_filter_ids(self):
        eval_filter, train_filter, golden_filter = generate_spec_eval_filter_ids(
            "eval_test", "train_test", "golden_test"
        )
        assert eval_filter == "tag::eval_test"
        assert train_filter == "tag::train_test"
        assert golden_filter == "tag::golden_test"

    def test_generates_filter_ids_with_tag_prefix(self):
        eval_filter, train_filter, golden_filter = generate_spec_eval_filter_ids(
            "my_tag", "other_tag", "third_tag"
        )
        assert eval_filter.startswith("tag::")
        assert train_filter.startswith("tag::")
        assert golden_filter.startswith("tag::")



from unittest.mock import patch

import pytest
from fastapi import FastAPI
from fastapi.testclient import TestClient
from kiln_ai.datamodel import Project, Task
from kiln_ai.datamodel.datamodel_enums import Priority
from kiln_ai.datamodel.eval import Eval, EvalOutputScore, TaskOutputRatingType
from kiln_ai.datamodel.spec import Spec, SpecStatus, TaskSample
from kiln_ai.datamodel.spec_properties import (
    DesiredBehaviourProperties,
    HallucinationsProperties,
    SpecType,
    ToneProperties,
    ToxicityProperties,
)

from kiln_server.custom_errors import connect_custom_errors
from kiln_server.spec_api import connect_spec_api


@pytest.fixture
def app():
    app = FastAPI()
    connect_spec_api(app)
    connect_custom_errors(app)
    return app


@pytest.fixture
def client(app):
    return TestClient(app)


@pytest.fixture
def project_and_task(tmp_path):
    project_path = tmp_path / "test_project" / "project.kiln"
    project_path.parent.mkdir()

    project = Project(name="Test Project", path=project_path)
    project.save_to_file()
    task = Task(
        name="Test Task",
        instruction="This is a test instruction",
        description="This is a test task",
        parent=project,
    )
    task.save_to_file()

    return project, task


def create_task_sample_dict():
    """Helper to create a valid task sample for API tests."""
    return {
        "input": "Example input",
        "output": "Example output",
    }


def create_tone_properties_dict():
    """Helper to create valid tone properties for API tests."""
    return {
        "spec_type": SpecType.tone.value,
        "core_requirement": "Test instruction",
        "tone_description": "Professional and friendly",
    }


def create_toxicity_properties_dict():
    """Helper to create valid toxicity properties for API tests."""
    return {
        "spec_type": SpecType.toxicity.value,
        "core_requirement": "Test instruction",
        "toxicity_examples": "Example toxicity content",
    }


def create_reference_answer_accuracy_properties_dict():
    """Helper to create valid reference answer accuracy properties for API tests."""
    return {
        "spec_type": SpecType.reference_answer_accuracy.value,
        "core_requirement": "Test instruction",
        "reference_answer_accuracy_description": "Must match reference",
        "accurate_examples": "Accurate example",
        "inaccurate_examples": "Inaccurate example",
    }


@pytest.fixture
def sample_tone_properties():
    """Fixture for creating complete ToneProperties objects for direct Spec creation."""
    return ToneProperties(
        spec_type=SpecType.tone,
        core_requirement="Test instruction",
        tone_description="Professional and friendly",
    )


@pytest.fixture
def sample_hallucinations_properties():
    """Fixture for creating complete HallucinationsProperties objects."""
    return HallucinationsProperties(
        spec_type=SpecType.hallucinations,
        core_requirement="Test instruction",
        hallucinations_examples="Example hallucination",
    )


@pytest.fixture
def sample_toxicity_properties():
    """Fixture for creating complete ToxicityProperties objects."""
    return ToxicityProperties(
        spec_type=SpecType.toxicity,
        core_requirement="Test instruction",
        toxicity_examples="Example toxicity",
    )


@pytest.fixture
def sample_desired_behaviour_properties():
    """Fixture for creating complete DesiredBehaviourProperties objects."""
    return DesiredBehaviourProperties(
        spec_type=SpecType.desired_behaviour,
        core_requirement="Test instruction",
        desired_behaviour_description="Avoid toxic content",
        correct_behaviour_examples=None,
        incorrect_behaviour_examples=None,
    )


def test_create_spec_success(client, project_and_task):
    project, task = project_and_task

    spec_data = {
        "name": "Test Spec",
        "definition": "The system should always respond politely",
        "priority": Priority.p1,
        "status": SpecStatus.active.value,
        "tags": ["test", "important"],
        "properties": create_tone_properties_dict(),
        "task_sample": create_task_sample_dict(),
    }

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.post(
            f"/api/projects/{project.id}/tasks/{task.id}/spec", json=spec_data
        )

    assert response.status_code == 200
    res = response.json()
    assert res["name"] == "Test Spec"
    assert res["definition"] == "The system should always respond politely"
    assert res["properties"]["spec_type"] == SpecType.tone.value
    assert res["priority"] == 1
    assert res["status"] == "active"
    assert res["tags"] == ["test", "important"]
    # Verify eval_id was auto-generated
    assert res["eval_id"] is not None
    assert len(res["eval_id"]) > 0

    # Check that the spec was saved to the task/file
    specs = task.specs()
    assert len(specs) == 1
    assert specs[0].name == "Test Spec"
    assert specs[0].definition == "The system should always respond politely"
    assert specs[0].properties["spec_type"] == SpecType.tone
    assert specs[0].priority == Priority.p1
    assert specs[0].status == SpecStatus.active

    # Check that an eval was created with correct filter IDs
    evals = task.evals()
    assert len(evals) == 1
    assert evals[0].name == "Test Spec"
    assert evals[0].id == res["eval_id"]
    assert evals[0].eval_set_filter_id == "tag::eval_test_spec"
    assert evals[0].train_set_filter_id == "tag::train_test_spec"
    assert evals[0].eval_configs_filter_id == "tag::eval_golden_test_spec"


def test_create_spec_minimal(client, project_and_task):
    """Test creating a spec with minimal required fields (name, definition, properties only)."""
    project, task = project_and_task

    spec_data = {
        "name": "Minimal Spec",
        "definition": "No toxic content allowed",
        "properties": create_toxicity_properties_dict(),
    }

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.post(
            f"/api/projects/{project.id}/tasks/{task.id}/spec", json=spec_data
        )

    assert response.status_code == 200
    res = response.json()
    assert res["name"] == "Minimal Spec"
    assert res["definition"] == "No toxic content allowed"
    assert res["properties"]["spec_type"] == SpecType.toxicity.value
    # Check defaults were applied
    assert res["priority"] == 1  # Priority.p1
    assert res["status"] == "active"
    assert res["tags"] == []
    # Eval should be auto-created
    assert res["eval_id"] is not None


def test_create_spec_task_not_found(client):
    spec_data = {
        "name": "Test Spec",
        "definition": "System should behave correctly",
        "properties": create_tone_properties_dict(),
    }

    response = client.post(
        "/api/projects/project-id/tasks/fake-task-id/spec", json=spec_data
    )
    assert response.status_code == 404


def test_get_specs_success(
    client, project_and_task, sample_tone_properties, sample_toxicity_properties
):
    project, task = project_and_task

    spec1 = Spec(
        name="Spec 1",
        definition="System should respond appropriately",
        properties=sample_tone_properties,
        eval_id="test_eval_id_1",
        parent=task,
    )
    spec1.save_to_file()

    spec2 = Spec(
        name="Spec 2",
        definition="No toxic responses",
        priority=Priority.p3,
        properties=sample_toxicity_properties,
        eval_id="test_eval_id_2",
        parent=task,
    )
    spec2.save_to_file()

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.get(f"/api/projects/{project.id}/tasks/{task.id}/specs")

    assert response.status_code == 200
    res = response.json()
    assert isinstance(res, list)
    assert len(res) == 2
    spec_names = {spec["name"] for spec in res}
    assert spec_names == {"Spec 1", "Spec 2"}


def test_get_specs_empty(client, project_and_task):
    project, task = project_and_task

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.get(f"/api/projects/{project.id}/tasks/{task.id}/specs")

    assert response.status_code == 200
    res = response.json()
    assert isinstance(res, list)
    assert len(res) == 0


def test_get_specs_task_not_found(client):
    response = client.get("/api/projects/project-id/tasks/fake-task-id/specs")
    assert response.status_code == 404


def test_get_spec_success(client, project_and_task, sample_hallucinations_properties):
    project, task = project_and_task

    spec = Spec(
        name="Test Spec",
        definition="System should not hallucinate facts",
        priority=Priority.p2,
        status=SpecStatus.active,
        tags=["validation", "safety"],
        properties=sample_hallucinations_properties,
        eval_id="test_eval_id",
        parent=task,
    )
    spec.save_to_file()

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.get(
            f"/api/projects/{project.id}/tasks/{task.id}/specs/{spec.id}"
        )

    assert response.status_code == 200
    res = response.json()
    assert res["name"] == "Test Spec"
    assert res["definition"] == "System should not hallucinate facts"
    assert res["properties"]["spec_type"] == SpecType.hallucinations.value
    assert res["priority"] == 2
    assert res["status"] == "active"
    assert res["tags"] == ["validation", "safety"]


def test_get_spec_not_found(client, project_and_task):
    project, task = project_and_task

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.get(
            f"/api/projects/{project.id}/tasks/{task.id}/specs/nonexistent_id"
        )

    assert response.status_code == 404
    assert "Spec not found" in response.json()["message"]


def test_update_spec_success(client, project_and_task, sample_tone_properties):
    project, task = project_and_task

    spec = Spec(
        name="Original Name",
        definition="Original definition",
        priority=Priority.p3,
        status=SpecStatus.active,
        tags=["old_tag"],
        properties=sample_tone_properties,
        eval_id="test_eval_id",
        parent=task,
    )
    spec.save_to_file()

    update_data = {
        "name": "Updated Name",
    }

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.patch(
            f"/api/projects/{project.id}/tasks/{task.id}/specs/{spec.id}",
            json=update_data,
        )

    assert response.status_code == 200
    res = response.json()
    assert res["name"] == "Updated Name"
    # Verify other fields remain unchanged
    assert res["definition"] == "Original definition"
    assert res["priority"] == 3
    assert res["status"] == "active"
    assert res["tags"] == ["old_tag"]
    assert res["properties"]["spec_type"] == SpecType.tone.value

    # Verify the spec was updated in the task/file
    updated_spec = next((s for s in task.specs() if s.id == spec.id), None)
    assert updated_spec is not None
    assert updated_spec.name == "Updated Name"
    # Verify other fields remain unchanged
    assert updated_spec.definition == "Original definition"
    assert updated_spec.priority == Priority.p3
    assert updated_spec.status == SpecStatus.active
    assert updated_spec.tags == ["old_tag"]


def test_update_spec_with_existing_eval_id(
    client, project_and_task, sample_toxicity_properties
):
    """Test that updating a spec's name doesn't affect its eval_id."""
    project, task = project_and_task

    spec = Spec(
        name="Original Name",
        definition="Original definition",
        priority=Priority.p2,
        status=SpecStatus.active,
        tags=["old_tag"],
        eval_id="original_eval_id",
        properties=sample_toxicity_properties,
        parent=task,
    )
    spec.save_to_file()

    update_data = {
        "name": "Updated Name",
    }

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.patch(
            f"/api/projects/{project.id}/tasks/{task.id}/specs/{spec.id}",
            json=update_data,
        )

    assert response.status_code == 200
    res = response.json()
    assert res["name"] == "Updated Name"
    # Verify other fields including eval_id remain unchanged
    assert res["definition"] == "Original definition"
    assert res["properties"]["spec_type"] == SpecType.toxicity.value
    assert res["priority"] == 2
    assert res["status"] == "active"
    assert res["eval_id"] == "original_eval_id"


def test_update_spec_tags_only(client, project_and_task, sample_tone_properties):
    """Test updating only tags field (save_tags use case)."""
    project, task = project_and_task

    spec = Spec(
        name="Original Name",
        definition="Original definition",
        priority=Priority.p3,
        status=SpecStatus.active,
        tags=["old_tag"],
        eval_id="original_eval_id",
        properties=sample_tone_properties,
        parent=task,
    )
    spec.save_to_file()

    # Simulate save_tags function sending only tags
    update_data = {
        "tags": ["new_tag", "updated_tag"],
    }

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.patch(
            f"/api/projects/{project.id}/tasks/{task.id}/specs/{spec.id}",
            json=update_data,
        )

    assert response.status_code == 200
    res = response.json()
    assert res["tags"] == ["new_tag", "updated_tag"]
    # Verify other fields remain unchanged
    assert res["name"] == "Original Name"
    assert res["definition"] == "Original definition"
    assert res["priority"] == 3
    assert res["status"] == "active"
    assert res["eval_id"] == "original_eval_id"

    # Verify the change persisted
    updated_spec = next((s for s in task.specs() if s.id == spec.id), None)
    assert updated_spec is not None
    assert updated_spec.tags == ["new_tag", "updated_tag"]


def test_update_spec_status_only(client, project_and_task, sample_tone_properties):
    """Test updating only status field (archive use case)."""
    project, task = project_and_task

    spec = Spec(
        name="Test Spec",
        definition="Test definition",
        priority=Priority.p2,
        status=SpecStatus.active,
        tags=["test"],
        properties=sample_tone_properties,
        eval_id="test_eval_id",
        parent=task,
    )
    spec.save_to_file()

    # Update only status to archived
    update_data = {
        "status": SpecStatus.archived.value,
    }

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.patch(
            f"/api/projects/{project.id}/tasks/{task.id}/specs/{spec.id}",
            json=update_data,
        )

    assert response.status_code == 200
    res = response.json()
    assert res["status"] == "archived"
    # Verify other fields remain unchanged
    assert res["name"] == "Test Spec"
    assert res["definition"] == "Test definition"
    assert res["priority"] == 2
    assert res["tags"] == ["test"]

    # Verify the change persisted
    updated_spec = next((s for s in task.specs() if s.id == spec.id), None)
    assert updated_spec is not None
    assert updated_spec.status == SpecStatus.archived


def test_update_spec_not_found(client, project_and_task):
    project, task = project_and_task

    update_data = {
        "name": "Updated Name",
    }

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.patch(
            f"/api/projects/{project.id}/tasks/{task.id}/specs/nonexistent_id",
            json=update_data,
        )

    assert response.status_code == 404
    assert "Spec not found" in response.json()["message"]


def test_create_spec_creates_eval_with_correct_template(client, project_and_task):
    """Test that creating a spec auto-creates an eval with the correct template for the spec type."""
    project, task = project_and_task

    spec_data = {
        "name": "Reference Answer Spec",
        "definition": "Answers must match reference answers",
        "properties": create_reference_answer_accuracy_properties_dict(),
    }

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.post(
            f"/api/projects/{project.id}/tasks/{task.id}/spec", json=spec_data
        )

    assert response.status_code == 200
    res = response.json()
    assert res["eval_id"] is not None

    # Verify eval was created with correct properties
    evals = task.evals()
    assert len(evals) == 1
    assert evals[0].name == "Reference Answer Spec"
    # Reference answer spec should use 'rag' template
    assert evals[0].template == "rag"

    specs = task.specs()
    assert len(specs) == 1
    assert specs[0].eval_id == evals[0].id


def test_create_spec_with_properties(client, project_and_task):
    project, task = project_and_task

    spec_data = {
        "name": "Desired Behaviour Spec",
        "definition": "System should avoid toxic language",
        "properties": {
            "spec_type": SpecType.desired_behaviour.value,
            "core_requirement": "Test instruction",
            "desired_behaviour_description": "Avoid toxic language and offensive content",
            "incorrect_behaviour_examples": "Example 1: Don't use slurs\nExample 2: Don't be rude",
        },
    }

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.post(
            f"/api/projects/{project.id}/tasks/{task.id}/spec", json=spec_data
        )

    assert response.status_code == 200
    res = response.json()
    assert res["properties"] is not None
    assert res["properties"]["spec_type"] == SpecType.desired_behaviour.value
    assert (
        res["properties"]["desired_behaviour_description"]
        == "Avoid toxic language and offensive content"
    )
    assert (
        res["properties"]["incorrect_behaviour_examples"]
        == "Example 1: Don't use slurs\nExample 2: Don't be rude"
    )

    specs = task.specs()
    assert len(specs) == 1
    assert specs[0].properties is not None
    assert specs[0].properties["spec_type"] == SpecType.desired_behaviour


def test_create_spec_with_archived_status(client, project_and_task):
    """Test creating a spec with archived status."""
    project, task = project_and_task

    spec_data = {
        "name": "Archived Spec",
        "definition": "This spec is archived",
        "status": SpecStatus.archived.value,
        "properties": create_tone_properties_dict(),
    }

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.post(
            f"/api/projects/{project.id}/tasks/{task.id}/spec", json=spec_data
        )

    assert response.status_code == 200
    res = response.json()
    assert res["name"] == "Archived Spec"
    assert res["status"] == "archived"

    specs = task.specs()
    assert len(specs) == 1
    assert specs[0].status == SpecStatus.archived


def test_get_spec_with_archived_status(
    client, project_and_task, sample_tone_properties
):
    """Test getting a spec with archived status."""
    project, task = project_and_task

    spec = Spec(
        name="Archived Spec",
        definition="This spec is archived",
        status=SpecStatus.archived,
        properties=sample_tone_properties,
        eval_id="test_eval_id",
        parent=task,
    )
    spec.save_to_file()

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.get(
            f"/api/projects/{project.id}/tasks/{task.id}/specs/{spec.id}"
        )

    assert response.status_code == 200
    res = response.json()
    assert res["name"] == "Archived Spec"
    assert res["status"] == "archived"


# Validation error tests (422 responses)


def test_create_spec_missing_name(client, project_and_task):
    project, task = project_and_task

    spec_data = {
        "definition": "The system should always respond politely",
        "properties": create_tone_properties_dict(),
    }

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.post(
            f"/api/projects/{project.id}/tasks/{task.id}/spec", json=spec_data
        )

    assert response.status_code == 422
    res = response.json()
    assert "source_errors" in res
    assert any(
        error["loc"] == ["body", "name"] and error["type"] == "missing"
        for error in res["source_errors"]
    )


def test_create_spec_missing_definition(client, project_and_task):
    project, task = project_and_task

    spec_data = {
        "name": "Test Spec",
        "properties": create_tone_properties_dict(),
    }

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.post(
            f"/api/projects/{project.id}/tasks/{task.id}/spec", json=spec_data
        )

    assert response.status_code == 422
    res = response.json()
    assert "source_errors" in res
    assert any(
        error["loc"] == ["body", "definition"] and error["type"] == "missing"
        for error in res["source_errors"]
    )


def test_create_spec_missing_properties(client, project_and_task):
    project, task = project_and_task

    spec_data = {
        "name": "Test Spec",
        "definition": "The system should always respond politely",
        "properties": None,
    }

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.post(
            f"/api/projects/{project.id}/tasks/{task.id}/spec", json=spec_data
        )

    assert response.status_code == 422
    res = response.json()
    assert "source_errors" in res
    # Properties is required now, so missing it gives model_attributes_type error
    assert any(error["loc"] == ["body", "properties"] for error in res["source_errors"])


def test_create_spec_priority_default(client, project_and_task):
    """Test that priority defaults to p1 when not provided."""
    project, task = project_and_task

    spec_data = {
        "name": "Test Spec",
        "definition": "The system should always respond politely",
        "properties": create_tone_properties_dict(),
    }

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.post(
            f"/api/projects/{project.id}/tasks/{task.id}/spec", json=spec_data
        )

    assert response.status_code == 200
    res = response.json()
    assert res["priority"] == 1  # Priority.p1


def test_create_spec_status_default(client, project_and_task):
    """Test that status defaults to active when not provided."""
    project, task = project_and_task

    spec_data = {
        "name": "Test Spec",
        "definition": "The system should always respond politely",
        "properties": create_tone_properties_dict(),
    }

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.post(
            f"/api/projects/{project.id}/tasks/{task.id}/spec", json=spec_data
        )

    assert response.status_code == 200
    res = response.json()
    assert res["status"] == "active"


def test_create_spec_tags_default(client, project_and_task):
    """Test that tags defaults to empty list when not provided."""
    project, task = project_and_task

    spec_data = {
        "name": "Test Spec",
        "definition": "The system should always respond politely",
        "properties": create_tone_properties_dict(),
    }

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.post(
            f"/api/projects/{project.id}/tasks/{task.id}/spec", json=spec_data
        )

    assert response.status_code == 200
    res = response.json()
    assert res["tags"] == []


def test_create_spec_invalid_spec_type_in_properties(client, project_and_task):
    project, task = project_and_task

    spec_data = {
        "name": "Test Spec",
        "definition": "The system should always respond politely",
        "properties": {"spec_type": "invalid_type_value"},
    }

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.post(
            f"/api/projects/{project.id}/tasks/{task.id}/spec", json=spec_data
        )

    assert response.status_code == 422
    res = response.json()
    assert "source_errors" in res
    # Check that properties validation failed with invalid spec_type
    assert any(
        "properties" in str(error.get("loc", [])) for error in res["source_errors"]
    )


def test_create_spec_invalid_priority_enum(client, project_and_task):
    project, task = project_and_task

    spec_data = {
        "name": "Test Spec",
        "definition": "The system should always respond politely",
        "priority": "p99",
        "properties": create_tone_properties_dict(),
    }

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.post(
            f"/api/projects/{project.id}/tasks/{task.id}/spec", json=spec_data
        )

    assert response.status_code == 422
    res = response.json()
    assert "source_errors" in res
    assert any(
        error["loc"] == ["body", "priority"] and error["type"] == "enum"
        for error in res["source_errors"]
    )


def test_create_spec_invalid_status_enum(client, project_and_task):
    project, task = project_and_task

    spec_data = {
        "name": "Test Spec",
        "definition": "The system should always respond politely",
        "status": "pending",
        "properties": create_tone_properties_dict(),
    }

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.post(
            f"/api/projects/{project.id}/tasks/{task.id}/spec", json=spec_data
        )

    assert response.status_code == 422
    res = response.json()
    assert "source_errors" in res
    assert any(
        error["loc"] == ["body", "status"] and error["type"] == "enum"
        for error in res["source_errors"]
    )


def test_create_spec_invalid_name_type(client, project_and_task):
    project, task = project_and_task

    spec_data = {
        "name": 12345,
        "definition": "The system should always respond politely",
        "properties": create_tone_properties_dict(),
    }

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.post(
            f"/api/projects/{project.id}/tasks/{task.id}/spec", json=spec_data
        )

    assert response.status_code == 422
    res = response.json()
    assert "source_errors" in res
    assert any(error["loc"] == ["body", "name"] for error in res["source_errors"])


def test_create_spec_invalid_tags_type(client, project_and_task):
    project, task = project_and_task

    spec_data = {
        "name": "Test Spec",
        "definition": "The system should always respond politely",
        "tags": "not_a_list",
        "properties": create_tone_properties_dict(),
    }

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.post(
            f"/api/projects/{project.id}/tasks/{task.id}/spec", json=spec_data
        )

    assert response.status_code == 422
    res = response.json()
    assert "source_errors" in res
    assert any(error["loc"] == ["body", "tags"] for error in res["source_errors"])


def test_create_spec_empty_string_in_tags(client, project_and_task):
    project, task = project_and_task

    spec_data = {
        "name": "Test Spec",
        "definition": "The system should always respond politely",
        "tags": [""],
        "properties": create_tone_properties_dict(),
    }

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.post(
            f"/api/projects/{project.id}/tasks/{task.id}/spec", json=spec_data
        )

    assert response.status_code == 422
    res = response.json()
    assert "source_errors" in res
    assert any(
        "empty strings" in error.get("msg", "").lower()
        for error in res["source_errors"]
    )


def test_create_spec_tag_with_space(client, project_and_task):
    project, task = project_and_task

    spec_data = {
        "name": "Test Spec",
        "definition": "The system should always respond politely",
        "tags": ["tag with space"],
        "properties": create_tone_properties_dict(),
    }

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.post(
            f"/api/projects/{project.id}/tasks/{task.id}/spec", json=spec_data
        )

    assert response.status_code == 422
    res = response.json()
    assert "source_errors" in res
    assert any(
        (
            "spaces" in error.get("msg", "").lower()
            or "underscores" in error.get("msg", "").lower()
        )
        for error in res["source_errors"]
    )


def test_update_spec_invalid_name_type(
    client, project_and_task, sample_tone_properties
):
    """Test that updating a spec with invalid name type fails."""
    project, task = project_and_task

    spec = Spec(
        name="Test Spec",
        definition="System should behave correctly",
        properties=sample_tone_properties,
        eval_id="test_eval_id",
        parent=task,
    )
    spec.save_to_file()

    update_data = {
        "name": 12345,
    }

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.patch(
            f"/api/projects/{project.id}/tasks/{task.id}/specs/{spec.id}",
            json=update_data,
        )

    assert response.status_code == 422
    res = response.json()
    assert "source_errors" in res
    assert any(error["loc"] == ["body", "name"] for error in res["source_errors"])


def test_create_spec_with_empty_tool_function_name(client, project_and_task):
    project, task = project_and_task

    spec_data = {
        "name": "Tool Use Spec",
        "definition": "Tool use validation test",
        "properties": {
            "spec_type": "appropriate_tool_use",
            "core_requirement": "Test instruction",
            "tool_id": "test_tool",
            "tool_function_name": "",
            "tool_use_guidelines": "Use this tool when needed",
            "appropriate_tool_use_examples": "examples",
            "inappropriate_tool_use_examples": "examples",
        },
    }

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.post(
            f"/api/projects/{project.id}/tasks/{task.id}/spec", json=spec_data
        )

    assert response.status_code == 422
    res = response.json()
    assert "source_errors" in res
    assert any(
        "tool_function_name" in error.get("msg", "").lower()
        for error in res["source_errors"]
    )


def test_create_spec_with_empty_tool_use_guidelines(client, project_and_task):
    project, task = project_and_task

    spec_data = {
        "name": "Tool Use Spec",
        "definition": "Tool use validation test",
        "properties": {
            "spec_type": "appropriate_tool_use",
            "core_requirement": "Test instruction",
            "tool_id": "test_tool",
            "tool_function_name": "test_tool_function",
            "tool_use_guidelines": "",
            "appropriate_tool_use_examples": "examples",
            "inappropriate_tool_use_examples": "examples",
        },
    }

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.post(
            f"/api/projects/{project.id}/tasks/{task.id}/spec", json=spec_data
        )

    assert response.status_code == 422
    res = response.json()
    assert "source_errors" in res
    assert any(
        "tool_use_guidelines" in error.get("msg", "").lower()
        for error in res["source_errors"]
    )


def test_create_spec_with_empty_behavior_description(client, project_and_task):
    project, task = project_and_task

    spec_data = {
        "name": "Desired Behaviour Spec",
        "definition": "Desired behaviour validation test",
        "properties": {
            "spec_type": "desired_behaviour",
            "core_requirement": "Test instruction",
            "desired_behaviour_description": "",
            "incorrect_behaviour_examples": "Example 1: Don't do this",
        },
    }

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.post(
            f"/api/projects/{project.id}/tasks/{task.id}/spec", json=spec_data
        )

    assert response.status_code == 422
    res = response.json()
    assert "source_errors" in res
    assert any(
        "desired_behaviour_description" in error.get("msg", "").lower()
        for error in res["source_errors"]
    )


def test_create_spec_with_empty_core_requirement(client, project_and_task):
    project, task = project_and_task

    spec_data = {
        "name": "Tone Spec",
        "definition": "Tone validation test",
        "properties": {
            "spec_type": "tone",
            "core_requirement": "",
            "tone_description": "Professional and friendly",
        },
    }

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.post(
            f"/api/projects/{project.id}/tasks/{task.id}/spec", json=spec_data
        )

    assert response.status_code == 422
    res = response.json()
    assert "source_errors" in res
    assert any(
        "core_requirement" in error.get("msg", "").lower()
        for error in res["source_errors"]
    )


def test_delete_spec_success(client, project_and_task, sample_tone_properties):
    project, task = project_and_task

    spec = Spec(
        name="Test Spec",
        definition="System should behave correctly",
        properties=sample_tone_properties,
        eval_id="test_eval_id",
        parent=task,
    )
    spec.save_to_file()

    specs = task.specs()
    assert len(specs) == 1

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.delete(
            f"/api/projects/{project.id}/tasks/{task.id}/specs/{spec.id}"
        )

    assert response.status_code == 200

    specs = task.specs()
    assert len(specs) == 0


def test_delete_spec_not_found(client, project_and_task):
    project, task = project_and_task

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.delete(
            f"/api/projects/{project.id}/tasks/{task.id}/specs/nonexistent_id"
        )

    assert response.status_code == 404
    assert "Spec not found" in response.json()["message"]


def test_delete_spec_with_associated_eval(
    client, project_and_task, sample_tone_properties
):
    """Test that deleting a spec also deletes its associated eval."""
    project, task = project_and_task

    # Create an eval with required fields (using 'rag' template to avoid needing eval_configs_filter_id)
    eval = Eval(
        name="Test Eval",
        description="Test eval description",
        template="rag",
        eval_set_filter_id="tag::test_eval",
        output_scores=[
            EvalOutputScore(
                name="Quality",
                type=TaskOutputRatingType.five_star,
            )
        ],
        parent=task,
    )
    eval.save_to_file()

    # Create a spec with the eval_id
    spec = Spec(
        name="Test Spec",
        definition="System should behave correctly",
        properties=sample_tone_properties,
        eval_id=eval.id,
        parent=task,
    )
    spec.save_to_file()

    # Verify both exist
    specs = task.specs()
    evals = task.evals()
    assert len(specs) == 1
    assert len(evals) == 1

    # Delete the spec
    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.delete(
            f"/api/projects/{project.id}/tasks/{task.id}/specs/{spec.id}"
        )

    assert response.status_code == 200

    # Verify both spec and eval are deleted
    specs = task.specs()
    evals = task.evals()
    assert len(specs) == 0
    assert len(evals) == 0


def test_delete_spec_without_associated_eval(
    client, project_and_task, sample_tone_properties
):
    """Test that deleting a spec with a non-existent eval works correctly."""
    project, task = project_and_task

    spec = Spec(
        name="Test Spec",
        definition="System should behave correctly",
        properties=sample_tone_properties,
        eval_id="nonexistent_eval_id",
        parent=task,
    )
    spec.save_to_file()

    specs = task.specs()
    assert len(specs) == 1

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.delete(
            f"/api/projects/{project.id}/tasks/{task.id}/specs/{spec.id}"
        )

    assert response.status_code == 200

    specs = task.specs()
    assert len(specs) == 0


def test_create_spec_with_task_sample(client, project_and_task):
    """Test creating a spec with a task sample."""
    project, task = project_and_task

    spec_data = {
        "name": "Spec With Sample",
        "definition": "Test spec with task sample",
        "properties": create_tone_properties_dict(),
        "task_sample": {
            "input": "What is the capital of France?",
            "output": "The capital of France is Paris.",
        },
    }

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.post(
            f"/api/projects/{project.id}/tasks/{task.id}/spec", json=spec_data
        )

    assert response.status_code == 200
    res = response.json()
    assert res["name"] == "Spec With Sample"
    assert res["task_sample"] is not None
    assert res["task_sample"]["input"] == "What is the capital of France?"
    assert res["task_sample"]["output"] == "The capital of France is Paris."

    specs = task.specs()
    assert len(specs) == 1
    assert specs[0].task_sample is not None
    assert specs[0].task_sample.input == "What is the capital of France?"
    assert specs[0].task_sample.output == "The capital of France is Paris."


def test_create_spec_without_task_sample(client, project_and_task):
    """Test creating a spec without a task sample (should default to None)."""
    project, task = project_and_task

    spec_data = {
        "name": "Spec Without Sample",
        "definition": "Test spec without task sample",
        "properties": create_tone_properties_dict(),
    }

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.post(
            f"/api/projects/{project.id}/tasks/{task.id}/spec", json=spec_data
        )

    assert response.status_code == 200
    res = response.json()
    assert res["name"] == "Spec Without Sample"
    assert res["task_sample"] is None

    specs = task.specs()
    assert len(specs) == 1
    assert specs[0].task_sample is None


def test_get_spec_with_task_sample(client, project_and_task, sample_tone_properties):
    """Test getting a spec with a task sample."""
    project, task = project_and_task

    sample = TaskSample(
        input="Example input for get test",
        output="Example output for get test",
    )
    spec = Spec(
        name="Test Spec With Sample",
        definition="System should behave correctly",
        properties=sample_tone_properties,
        eval_id="test_eval_id",
        task_sample=sample,
        parent=task,
    )
    spec.save_to_file()

    with patch("kiln_server.spec_api.task_from_id") as mock_task_from_id:
        mock_task_from_id.return_value = task
        response = client.get(
            f"/api/projects/{project.id}/tasks/{task.id}/specs/{spec.id}"
        )

    assert response.status_code == 200
    res = response.json()
    assert res["task_sample"] is not None
    assert res["task_sample"]["input"] == "Example input for get test"
    assert res["task_sample"]["output"] == "Example output for get test"



"""Runtime helpers for configuring and running the Kiln MCP server."""

from __future__ import annotations

import logging
from typing import Any, Literal, Sequence

from mcp.server.fastmcp import FastMCP
from mcp.types import TextContent

from .mcp_server_tool_utils import ToolContext

logger = logging.getLogger(__name__)

Transport = Literal["stdio", "sse", "streamable-http"]


def create_fastmcp_server(
    tool_contexts: Sequence[ToolContext],
    *,
    project_name: str | None,
    host: str,
    port: int,
    log_level: str,
    transport: Transport,
    mount_path: str | None,
) -> FastMCP:
    """Instantiate and configure a :class:`FastMCP` server for Kiln tools."""

    server_name = "Kiln MCP Server"
    if project_name:
        server_name = f"{server_name} ({project_name})"

    instructions = "Kiln MCP server exposing project tools."

    settings: dict[str, Any] = {
        "host": host,
        "port": port,
        "log_level": log_level.upper(),
    }
    if transport == "sse" and mount_path:
        settings["sse_path"] = mount_path
    if transport == "streamable-http" and mount_path:
        settings["streamable_http_path"] = mount_path

    server = FastMCP(name=server_name, instructions=instructions, **settings)

    tools_by_name = {context.definition.name: context for context in tool_contexts}

    @server._mcp_server.list_tools()
    async def _list_tools() -> list[Any]:
        return [context.definition for context in tool_contexts]

    @server._mcp_server.call_tool(validate_input=True)
    async def _call_tool(name: str, arguments: dict[str, Any] | None = None) -> Any:
        if arguments is None:
            arguments = {}
        if not isinstance(arguments, dict):
            raise ValueError("Tool arguments must be a JSON object.")

        context = tools_by_name.get(name)
        if context is None:
            raise ValueError(f"Unknown tool requested: {name}")

        try:
            result = await context.resolution.tool.run(**arguments)
        except Exception as exc:  # pragma: no cover - defensive logging
            logger.exception("Tool %s failed", name)
            raise ValueError(str(exc)) from exc

        return [TextContent(type="text", text=result.output)]

    return server


async def run_transport(
    server: FastMCP, transport: Transport, mount_path: str | None
) -> None:
    """Run the server using the selected transport."""

    if transport == "stdio":
        await server.run_stdio_async()
    elif transport == "sse":
        await server.run_sse_async(mount_path)
    elif transport == "streamable-http":
        await server.run_streamable_http_async()
    else:  # pragma: no cover - defensive branch
        raise ValueError(f"Unsupported transport: {transport}")
